/* YAD 0.03 (Yet Another Disassembler) (x) herm1t@vx.netlux.org, 2008
 *
 * This also might be called Less eXtended Disassembler Engine, because
 * it's primarily based on XDE v. 1.02 by Z0mbie, the stuff related to
 * operands parsing was omitted, but the table was regenerated to reflect
 * the changes in IA-32 since the last release of XDE. There were also
 * minor clean-ups in the code.
 *
 * The constants and the data structure is partially compatible, if you
 * gonna use this instead of XDE, you'll have to slightly change your code.
 *
 * The code was designed specially for the Linux.Lacrimae virus.
 *
 * ChangeLog:
 *
 * 2008-05-11	0.06 reduce table size, rfc1925 6a
 * 2008-05-11	0.05 merge main & float tables
 * 2008-05-11	0.05 get rid of ADDR2, DATA4 flags (suspicious)
 * 2008-05-11	0.05 dumb mistake in float table generation (missing bzero...)
 * 2008-05-09	0.05 minor clean-ups
 * 2008-05-08	0.04 C_STOP|C_REL for ljmp (0xea), C_STOP fpr lret/lretn (0xca/0xcb), iret (0xcf)
 *		YAD tester for unices by me and for Win32 by cqwerty
 * 2008-05-08	0.03 Didn't checked "extended" flags for error
 * 2008-05-08	0.02 FPU support
 * 2008-05-07	0.01 Initial release 
 */
#include "yad.h"
#include "yad_data.h"

int yad(uint8_t *opcode, yad_t *diza)
{
	uint8_t c, *p;
	uint32_t flags, datasize, addrsize, i, j;
	uint16_t *yad_table = (uint16_t*)(yad_data + YAD_VALUES_OFFSET);

	bzero(diza, sizeof(yad_t));
	p = opcode;	
	datasize = addrsize = 4;
	flags = 0;
	if (*(unsigned short*)p == 0x0000 || *(unsigned short*)p == 0xFFFF)
		flags |= C_BAD;
	/* prefixes */
	for (;;) {
		c = *p++;
#define	GET_PREFIX(cond,field,expr)	\
if (cond) { if (diza->field != 0) flags |= C_BAD; diza->field = c; expr; continue; }
		GET_PREFIX(c == 0x66, p_66, datasize = 2)
		GET_PREFIX(c == 0x67, p_67, addrsize = 2)
		GET_PREFIX((c == 0x26) || (c == 0x2E) || (c == 0x36) || (c == 0x3E) || (c == 0x64) || (c == 0x65), p_seg, )
		GET_PREFIX((c == 0xF2) || (c == 0xF3), p_rep, )
		GET_PREFIX(c == 0xf0, p_lock, )
		break;
	}
#undef	GET_PREFIX
	/* first opcode */
	diza->opcode = c;
	/* instruction flags by opcode */
	j = 0;
	if (c == 0x0f) {
		/* 2nd opcode */
		diza->opcode2 = c = *p++;
		j = 256 + c;
	} else
		j = c;
	if ((flags |= yad_table[yad_data[j]]) == C_ERROR)
		return 0;
	/* prefix groups */
	if (flags & C_PREGRP) {
		i = diza->p_rep == 0xf2 ? 3 : (diza->p_66 ? 2 : (diza->p_rep == 0xf3 ? 1 : 0));
		if ((flags = yad_table[yad_data[(flags & YAD_GROUP_MASK) + i]]) == C_ERROR)
			return 0;
	}
	/* parse ModRM and SIB */
	if (flags & (C_MODRM|C_GROUP)) {
		uint32_t mod, rm, reg;
		c = *p++;
		diza->modrm = c;
		mod = c >> 6;
		reg = (c >> 3) & 7;
		rm = c & 7;
		/* FPU */
		if ((c = diza->opcode) >= 0xd8 && c <= 0xdf) {
			c -= 0xd8;
			if (mod != 3) {
				i = reg;
			} else {
				c = ((c + 1) << 3) + reg;
				i = rm;
			}
			uint8_t *yad_float = &yad_data[YAD_FLOAT_OFFSET];
			if (yad_float[c] & (1 << i))
				return 0;
		} else
		/* groups */
		if ((flags & C_GROUP) && (flags = yad_table[yad_data[(flags & YAD_GROUP_MASK) + reg]]) == C_ERROR)
			return 0;
		if (mod != 3) {
			if (rm == 4) {
				c = *p++;
				flags |= C_SIB;
				diza->sib = c;
			}
			if (mod == 2 || (mod == 0 && (rm == 5 || (rm == 4 && (c & 7) == 5))))
				flags |= C_ADDR4;
			else if (mod == 1)
				flags |= C_ADDR1;
		}
	}
	diza->flags = flags;
#define	COPY_ARG(size,flag,var,save)	\
{ int a = size;				\
if (flags & flag)			\
	a += var;			\
for (i = 0; i < a; i++)			\
	*(&save + i) = *p++;		\
diza->var = a; }
	COPY_ARG((flags >> 3) & 5, C_ADDR67, addrsize, diza->addr1);
	COPY_ARG(flags & 3, C_DATA66, datasize, diza->data1);
	if ((diza->len = p - opcode) > 15)
		return 0;	
	return diza->len;
#undef	COPY_ARG
}

int yad_asm(uint8_t *opcode, yad_t *diza)
{
	uint8_t *p;
	int i;
#define	SET_PREFIX(var)	if (diza->var) *p++ = diza->var;
	p = opcode;
	SET_PREFIX(p_seg);
	SET_PREFIX(p_lock);
	SET_PREFIX(p_rep);
	SET_PREFIX(p_67);
	SET_PREFIX(p_66);
#undef	SET_PREFIX
	*p++ = diza->opcode;
	if (diza->opcode == 0x0f)
		*p++ = diza->opcode2;
	if (diza->flags & C_MODRM)
		*p++ = diza->modrm;
	if (diza->flags & C_SIB)
		*p++ = diza->sib;
	for (i = 0; i < diza->addrsize; i++)
		*p++ = *(&diza->addr1 + i);
	for (i = 0; i < diza->datasize; i++)
		*p++ = *(&diza->data1 + i);
	return p - opcode;
}
